/* $Id: key.c,v 1.1.1.1 2006/03/17 19:53:40 zicodxx Exp $ */
/*
 *
 * SDL keyboard input support
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include <SDL/SDL.h>

#include "event.h"
#include "error.h"
#include "key.h"
#include "timer.h"
#include "window.h"
#include "console.h"
#include "args.h"

static unsigned char Installed = 0;

//-------- Variable accessed by outside functions ---------
int			keyd_repeat = 0; // 1 = use repeats, 0 no repeats
volatile unsigned char 	keyd_last_pressed;
volatile unsigned char 	keyd_last_released;
volatile unsigned char	keyd_pressed[256];
fix64			keyd_time_when_last_pressed;
unsigned char		unicode_frame_buffer[KEY_BUFFER_SIZE] = { '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0' };

typedef struct keyboard	{
	ubyte state[256];
} keyboard;

static keyboard key_data;

typedef struct key_props {
	char *key_text;
	unsigned char ascii_value;
	SDLKey sym;
} key_props;

key_props key_properties[256] = { //Keycodes stripped so that the end-user will only see the dingoo keys.
{ "",       255,    -1                 }, // 0
{ "SELECT",    255,    SDLK_ESCAPE        },
{ "",      '1',    SDLK_1             },
{ "",      '2',    SDLK_2             },
{ "",      '3',    SDLK_3             },
{ "",      '4',    SDLK_4             },
{ "",      '5',    SDLK_5             },
{ "",      '6',    SDLK_6             },
{ "",      '7',    SDLK_7             },
{ "",      '8',    SDLK_8             },
{ "",      '9',    SDLK_9             }, // 10
{ "",      '0',    SDLK_0             },
{ "",      '-',    SDLK_MINUS         },
{ "",      '=',    SDLK_EQUALS        },
{ "R",   255,    SDLK_BACKSPACE     },
{ "L",    255,    SDLK_TAB           },
{ "",      'q',    SDLK_q             },
{ "",      'w',    SDLK_w             },
{ "",      'e',    SDLK_e             },
{ "",      'r',    SDLK_r             },
{ "",      't',    SDLK_t             }, // 20
{ "Y",      'y',    SDLK_y             },
{ "",      'u',    SDLK_u             },
{ "",      'i',    SDLK_i             },
{ "",      'o',    SDLK_o             },
{ "",      'p',    SDLK_p             },
{ "",      '[',    SDLK_LEFTBRACKET   },
{ "",      ']',    SDLK_RIGHTBRACKET  },
{ "START",  255,    SDLK_RETURN        },
{ "",  255,    SDLK_LCTRL         },
{ "A",      'a',    SDLK_a             }, // 30
{ "",      's',    SDLK_s             },
{ "",      'd',    SDLK_d             },
{ "",      'f',    SDLK_f             },
{ "",      'g',    SDLK_g             },
{ "",      'h',    SDLK_h             },
{ "",      'j',    SDLK_j             },
{ "",      'k',    SDLK_k             },
{ "",      'l',    SDLK_l             },
{ "",      ';',    SDLK_SEMICOLON     },
{ "",      '\'',   SDLK_QUOTE         }, // 40
{ "",      '`',    SDLK_BACKQUOTE     },
{ "",  255,    SDLK_LSHIFT        },
{ "",     '\\',   SDLK_BACKSLASH     },
{ "",      'z',    SDLK_z             },
{ "X",      'x',    SDLK_x             },
{ "",      'c',    SDLK_c             },
{ "",      'v',    SDLK_v             },
{ "B",      'b',    SDLK_b             },
{ "",      'n',    SDLK_n             },
{ "",      'm',    SDLK_m             }, // 50
{ "",      ',',    SDLK_COMMA         },
{ "",      '.',    SDLK_PERIOD        },
{ "",      '/',    SDLK_SLASH         },
{ "",  255,    SDLK_RSHIFT        },
{ "",   '*',    SDLK_KP_MULTIPLY   },
{ "",   255,    SDLK_LALT          },
{ "",    ' ',    SDLK_SPACE         },
{ "",  255,    SDLK_CAPSLOCK      },
{ "",     255,    SDLK_F1            },
{ "",     255,    SDLK_F2            }, // 60
{ "",     255,    SDLK_F3            },
{ "",     255,    SDLK_F4            },
{ "",     255,    SDLK_F5            },
{ "",     255,    SDLK_F6            },
{ "",     255,    SDLK_F7            },
{ "",     255,    SDLK_F8            },
{ "",     255,    SDLK_F9            },
{ "",    255,    SDLK_F10           },
{ "",  255,    SDLK_NUMLOCK       },
{ "",   255,    SDLK_SCROLLOCK     }, // 70
{ "",   255,    SDLK_KP7           },
{ "",   255,    SDLK_KP8           },
{ "",   255,    SDLK_KP9           },
{ "",   255,    SDLK_KP_MINUS      },
{ "SEL+Y",   255,    SDLK_KP4           },
{ "",   255,    SDLK_KP5           },
{ "",   255,    SDLK_KP6           },
{ "",   255,    SDLK_KP_PLUS       },
{ "SEL+A",   255,    SDLK_KP1           },
{ "SEL+B",   255,    SDLK_KP2           }, // 80
{ "SEL+X",   255,    SDLK_KP3           },
{ "",   255,    SDLK_KP0           },
{ "",   255,    SDLK_KP_PERIOD     },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",    255,    SDLK_F11           },
{ "",    255,    SDLK_F12           },
{ "",       255,    -1                 },
{ "",       255,    -1                 }, // 90
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "HOLD",  255,    SDLK_PAUSE         },
{ "",     255,    SDLK_WORLD_0       },
{ "",     255,    SDLK_WORLD_1       },
{ "",     255,    SDLK_WORLD_2       }, // 100
{ "",     255,    SDLK_WORLD_3       },
{ "",     255,    SDLK_WORLD_4       },
{ "",     255,    SDLK_WORLD_5       },
{ "",     255,    SDLK_WORLD_6       },
{ "",     255,    SDLK_WORLD_7       },
{ "",     255,    SDLK_WORLD_8       },
{ "",     255,    SDLK_WORLD_9       },
{ "",    255,    SDLK_WORLD_10      },
{ "",    255,    SDLK_WORLD_11      },
{ "",    255,    SDLK_WORLD_12      }, // 110
{ "",    255,    SDLK_WORLD_13      },
{ "",    255,    SDLK_WORLD_14      },
{ "",    255,    SDLK_WORLD_15      },
{ "",    255,    SDLK_WORLD_16      },
{ "",    255,    SDLK_WORLD_17      },
{ "",    255,    SDLK_WORLD_18      },
{ "",    255,    SDLK_WORLD_19      },
{ "",    255,    SDLK_WORLD_20      },
{ "",    255,    SDLK_WORLD_21      },
{ "",    255,    SDLK_WORLD_22      }, // 120
{ "",    255,    SDLK_WORLD_23      },
{ "",    255,    SDLK_WORLD_24      },
{ "",    255,    SDLK_WORLD_25      },
{ "",    255,    SDLK_WORLD_26      },
{ "",    255,    SDLK_WORLD_27      },
{ "",    255,    SDLK_WORLD_28      },
{ "",    255,    SDLK_WORLD_29      },
{ "",    255,    SDLK_WORLD_30      },
{ "",    255,    SDLK_WORLD_31      },
{ "",    255,    SDLK_WORLD_32      }, // 130
{ "",    255,    SDLK_WORLD_33      },
{ "",    255,    SDLK_WORLD_34      },
{ "",    255,    SDLK_WORLD_35      },
{ "",    255,    SDLK_WORLD_36      },
{ "",    255,    SDLK_WORLD_37      },
{ "",    255,    SDLK_WORLD_38      },
{ "",    255,    SDLK_WORLD_39      },
{ "",    255,    SDLK_WORLD_40      },
{ "",    255,    SDLK_WORLD_41      },
{ "",    255,    SDLK_WORLD_42      }, // 140
{ "",    255,    SDLK_WORLD_43      },
{ "",    255,    SDLK_WORLD_44      },
{ "",    255,    SDLK_WORLD_45      },
{ "",    255,    SDLK_WORLD_46      },
{ "",    255,    SDLK_WORLD_47      },
{ "",    255,    SDLK_WORLD_48      },
{ "",    255,    SDLK_WORLD_49      },
{ "",    255,    SDLK_WORLD_50      },
{ "",    255,    SDLK_WORLD_51      },
{ "",       255,    -1                 }, // 150
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",    255,    SDLK_KP_ENTER      },
{ "",  255,    SDLK_RCTRL         },
{ "",   255,    SDLK_LMETA         },
{ "",   255,    SDLK_RMETA         },
{ "",       255,    -1                 }, // 160
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 }, // 170
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 }, // 180
{ "",   255,    SDLK_KP_DIVIDE     },
{ "",       255,    -1                 },
{ "",  255,    SDLK_PRINT         },
{ "",   255,    SDLK_RALT          },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 }, // 190
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",       255,    -1                 },
{ "",   255,    SDLK_HOME          },
{ "UP",     255,    SDLK_UP            }, // 200
{ "",   255,    SDLK_PAGEUP        },
{ "",       255,    -1                 },
{ "LEFT",   255,    SDLK_LEFT          },
{ "",       255,    -1                 },
{ "RIGHT",  255,    SDLK_RIGHT         },
{ "",       255,    -1                 },
{ "",    255,    SDLK_END           },
{ "DOWN",   255,    SDLK_DOWN          },
{ "",   255,    SDLK_PAGEDOWN      },
{ "",    255,    SDLK_INSERT        }, // 210
{ "",    255,    SDLK_DELETE        },
{ "",    255,    SDLK_WORLD_52      },
{ "",    255,    SDLK_WORLD_53      },
{ "",    255,    SDLK_WORLD_54      },
{ "",    255,    SDLK_WORLD_55      },
{ "",    255,    SDLK_WORLD_56      },
{ "",    255,    SDLK_WORLD_57      },
{ "",    255,    SDLK_WORLD_58      },
{ "",    255,    SDLK_WORLD_59      },
{ "",    255,    SDLK_WORLD_60      }, // 220
{ "",    255,    SDLK_WORLD_61      },
{ "",    255,    SDLK_WORLD_62      },
{ "",    255,    SDLK_WORLD_63      },
{ "",    255,    SDLK_WORLD_64      },
{ "",    255,    SDLK_WORLD_65      },
{ "",    255,    SDLK_WORLD_66      },
{ "",    255,    SDLK_WORLD_67      },
{ "",    255,    SDLK_WORLD_68      },
{ "",    255,    SDLK_WORLD_69      },
{ "",    255,    SDLK_WORLD_70      }, // 230
{ "",    255,    SDLK_WORLD_71      },
{ "",    255,    SDLK_WORLD_72      },
{ "",    255,    SDLK_WORLD_73      },
{ "",    255,    SDLK_WORLD_74      },
{ "",    255,    SDLK_WORLD_75      },
{ "",    255,    SDLK_WORLD_76      },
{ "",    255,    SDLK_WORLD_77      },
{ "",    255,    SDLK_WORLD_78      },
{ "",    255,    SDLK_WORLD_79      },
{ "",    255,    SDLK_WORLD_80      }, // 240
{ "",    255,    SDLK_WORLD_81      },
{ "",    255,    SDLK_WORLD_82      },
{ "",    255,    SDLK_WORLD_83      },
{ "",    255,    SDLK_WORLD_84      },
{ "",    255,    SDLK_WORLD_85      },
{ "",    255,    SDLK_WORLD_86      },
{ "",    255,    SDLK_WORLD_87      },
{ "",    255,    SDLK_WORLD_88      },
{ "",    255,    SDLK_WORLD_89      },
{ "",    255,    SDLK_WORLD_90      }, // 250
{ "",    255,    SDLK_WORLD_91      },
{ "",    255,    SDLK_WORLD_92      },
{ "",    255,    SDLK_WORLD_93      },
{ "",    255,    SDLK_WORLD_94      },
{ "",    255,    SDLK_WORLD_95      }, // 255
};

typedef struct d_event_keycommand
{
	event_type	type;	// EVENT_KEY_COMMAND/RELEASE
	int			keycode;
} d_event_keycommand;

char *key_text[256];

int key_ismodlck(int keycode)
{
	switch (keycode)
	{
		case KEY_LSHIFT:
		case KEY_RSHIFT:
		case KEY_LALT:
		case KEY_RALT:
		case KEY_LCTRL:
		case KEY_RCTRL:
		case KEY_LMETA:
		case KEY_RMETA:
			return KEY_ISMOD;
		case KEY_NUMLOCK:
		case KEY_SCROLLOCK:
		case KEY_CAPSLOCK:
			return KEY_ISLCK;
		default:
			return 0;
	}
}

int isEscaped = 0;

unsigned char key_ascii()
{
	static unsigned char unibuffer[KEY_BUFFER_SIZE] = { '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0' };
	int i=0, offset=0, count=0;
	
	offset=strlen((const char*)unibuffer);

	// move temporal chars from unicode_frame_buffer to empty space behind last unibuffer char (if any)
	for (i=offset; i < KEY_BUFFER_SIZE; i++)
		if (unicode_frame_buffer[count] != '\0')
		{
			unibuffer[i]=unicode_frame_buffer[count];
			unicode_frame_buffer[count]='\0';
			count++;
		}

	// unibuffer is not empty. store first char, remove it, shift all chars one step left and then print our char
	if (unibuffer[0] != '\0')
	{
		unsigned char retval = unibuffer[0];
		unsigned char unibuffer_shift[KEY_BUFFER_SIZE];
		memset(unibuffer_shift,'\0',sizeof(unsigned char)*KEY_BUFFER_SIZE);
		memcpy(unibuffer_shift,unibuffer+1,sizeof(unsigned char)*(KEY_BUFFER_SIZE-1));
		memcpy(unibuffer,unibuffer_shift,sizeof(unsigned char)*KEY_BUFFER_SIZE);
		return retval;
	}
	else
		return 255;
}

void key_handler(SDL_KeyboardEvent *event)
{
	int keycode, event_keysym=-1, key_state;

	if (event->keysym.sym == SDLK_ESCAPE)
	{
		isEscaped = (event->state == SDL_PRESSED) ? 1:0;
		return;
	}

	if (isEscaped) //Key Combo Mod :D
		switch (event->keysym.sym)
		{
			case SDLK_LCTRL:
				event->keysym.sym = SDLK_KP1;
			break;
			case SDLK_LALT :
				event->keysym.sym = SDLK_KP2;
			break;
			case SDLK_SPACE:
				event->keysym.sym = SDLK_KP3;
			break;
			case SDLK_LSHIFT:
				event->keysym.sym = SDLK_KP4;
			break;
			case SDLK_RETURN:
				event->keysym.sym = SDLK_ESCAPE;
			break;
			default:
			break;
		}
	if (!isEscaped) //Ugly menu hack, needed for quite some reasons, also, turns the developing easier, win-win
		switch (event->keysym.sym)
		{
			case SDLK_LCTRL:
				event->keysym.sym = SDLK_a;
			break;
			case SDLK_LALT :
				event->keysym.sym = SDLK_b;
			break;
			case SDLK_SPACE:
				event->keysym.sym = SDLK_x;
			break;
			case SDLK_LSHIFT:
				event->keysym.sym = SDLK_y;
			break;
			default:
			break;
		}
	// Read SDLK symbol and state
        event_keysym = event->keysym.sym;
        key_state = (event->state == SDL_PRESSED)?1:0;

	// fill the unicode frame-related unicode buffer 
	if (key_state && event->keysym.unicode > 31 && event->keysym.unicode < 255)
	{
		int i = 0;
		for (i = 0; i < KEY_BUFFER_SIZE; i++)
			if (unicode_frame_buffer[i] == '\0')
			{
				unicode_frame_buffer[i] = event->keysym.unicode;
				break;
			}
	}

	//=====================================================
	for (keycode = 255; keycode > 0; keycode--)
		if (key_properties[keycode].sym == event_keysym)
			break;

	if (keycode == 0)
		return;

	/* 
	 * process the key if:
	 * - it's a valid key AND
	 * - if the keystate has changed OR
	 * - key state same as last one and game accepts key repeats but keep out mod/lock keys
	 */
	if (key_state != keyd_pressed[keycode] || (keyd_repeat && !key_ismodlck(keycode)))
	{
		d_event_keycommand event;

		// now update the key props
		if (key_state) {
			keyd_last_pressed = keycode;
			keyd_pressed[keycode] = key_data.state[keycode] = 1;
		} else {
			keyd_pressed[keycode] = key_data.state[keycode] = 0;
		}

		if ( keyd_pressed[KEY_LSHIFT] || keyd_pressed[KEY_RSHIFT])
			keycode |= KEY_SHIFTED;
		if ( keyd_pressed[KEY_LALT] || keyd_pressed[KEY_RALT])
			keycode |= KEY_ALTED;
		if ( keyd_pressed[KEY_LCTRL] || keyd_pressed[KEY_RCTRL])
			keycode |= KEY_CTRLED;
		if ( keyd_pressed[KEY_DELETE] )
			keycode |= KEY_DEBUGGED;
		if ( keyd_pressed[KEY_LMETA] || keyd_pressed[KEY_RMETA])
			keycode |= KEY_METAED;

		// We allowed the key to be added to the queue for now,
		// because there are still input loops without associated windows
		event.type = key_state?EVENT_KEY_COMMAND:EVENT_KEY_RELEASE;
		event.keycode = keycode;
		con_printf(CON_DEBUG, "Sending event %s: %s %s %s %s %s %s\n",
				(key_state)                  ? "EVENT_KEY_COMMAND": "EVENT_KEY_RELEASE",
				(keycode & KEY_METAED)	? "META" : "",
				(keycode & KEY_DEBUGGED)	? "DEBUG" : "",
				(keycode & KEY_CTRLED)	? "CTRL" : "",
				(keycode & KEY_ALTED)	? "ALT" : "",
				(keycode & KEY_SHIFTED)	? "SHIFT" : "",
				key_properties[keycode & 0xff].key_text
				);
		event_send((d_event *)&event);
	}
}

void key_close()
{
      Installed = 0;
}

void key_init()
{
	int i;
	
	if (Installed) return;

	Installed=1;
	SDL_EnableUNICODE(1);
	key_toggle_repeat(1);

	keyd_time_when_last_pressed = timer_query();
	
	for(i=0; i<256; i++)
		key_text[i] = key_properties[i].key_text;
	  
	// Clear the keyboard array
	key_flush();

	isEscaped = 0;
}

void key_flush()
{
 	int i;
	Uint8 *keystate = SDL_GetKeyState(NULL);

	if (!Installed)
		key_init();

	//Clear the unicode buffer
	for (i=0; i<KEY_BUFFER_SIZE; i++ )
		unicode_frame_buffer[i] = '\0';

	for (i=0; i<256; i++ )	{
		if (key_ismodlck(i) == KEY_ISLCK && keystate[key_properties[i].sym] && !GameArg.CtlNoStickyKeys) // do not flush status of sticky keys
		{
			keyd_pressed[i] = 1;
			key_data.state[i] = 0;
		}
		else
		{
			keyd_pressed[i] = 0;
			key_data.state[i] = 1;
		}
	}
}

int event_key_get(d_event *event)
{
	Assert(event->type == EVENT_KEY_COMMAND || event->type == EVENT_KEY_RELEASE);
	return ((d_event_keycommand *)event)->keycode;
}

// same as above but without mod states
int event_key_get_raw(d_event *event)
{
	int keycode = ((d_event_keycommand *)event)->keycode;
	Assert(event->type == EVENT_KEY_COMMAND || event->type == EVENT_KEY_RELEASE);
	if ( keycode & KEY_SHIFTED ) keycode &= ~KEY_SHIFTED;
	if ( keycode & KEY_ALTED ) keycode &= ~KEY_ALTED;
	if ( keycode & KEY_CTRLED ) keycode &= ~KEY_CTRLED;
	if ( keycode & KEY_DEBUGGED ) keycode &= ~KEY_DEBUGGED;
	if ( keycode & KEY_METAED ) keycode &= ~KEY_METAED;
	return keycode;
}

void key_toggle_repeat(int enable)
{
	if (enable)
	{
		if (SDL_EnableKeyRepeat(KEY_REPEAT_DELAY, KEY_REPEAT_INTERVAL) == 0)
			keyd_repeat = 1;
	}
	else
	{
		SDL_EnableKeyRepeat(0, 0);
		keyd_repeat = 0;
	}
	key_flush();
}
